
/*
 * scclass.cxx
 *
 * Contains the move-finding and strategy functions of Chris Street's
 * Scrabble AI.
 *
 */

#undef USES_TIMERS

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "scrab.h"
#ifdef   USES_TIMERS
#include "timers.h"
#endif

bool use_feature;

#ifdef USES_TIMERS
extern volatile bool redraw_timer;
extern int timer, timer_x, timer_y;
#endif

#undef TEST_SCRAB

#include <iostream.h>

#define VOWEL(c) (c=='a' || c=='e' || c=='i' || c=='o' || c=='u')

struct AnaObject *create_new_AnaObject();
int compare_scores(void *, void *);

struct AnaObject anagrams;

#ifdef MEMORY_INTENSIVE

struct AnaObject *create_new_AnaObject() {
    struct AnaObject *a;
    int e;
    a=new struct AnaObject;
    for (e=0;e<26;++e)
        a->a[e]=NULL;
    a->is_word=false;
    return a;
}

#else

struct AnaObject *create_new_AnaObject() {
    struct AnaObject *a;
    a=new struct AnaObject;
    a->info=0;
    return a;
}

#endif

int compare_scores(const void *m1, const void *m2) {
    return ((struct Move *)(m2))->score*10 - ((struct Move *)(m1))->score*10 + ((struct Move *)(m2))->equity - ((struct Move *)(m1))->equity;
}

Dictionary::Dictionary(int max_vocab) {
        word=new char * [max_vocab];
        return;
}

Dictionary::~Dictionary() {
        delete [] word;
        return;
}

char *Dictionary::report_word(int which_one) {
     if (which_one>=nwords)
        return NULL;
     return word[which_one];
}

#ifdef MEMORY_INTENSIVE

int Dictionary::check_word(char *w) {
    AnaObject *a;
    int e=0;
    a=&anagrams;

    while (w[e]) {
          a=a->a[w[e++]-'a'];
          if (a==NULL)
             return -1;
    }

    if (a->is_word)
       return a->name;
    else
       return -1;
}

#else

int Dictionary::check_word(char *w) {
    AnaObject *a;
    int e=0,f;
    a=&anagrams;

    while (w[e]) {
          int i,j;
          j=0;
          i=(2 << (w[e]-'a'));
          if (!(a->info & i))
             return -1;
          i=1;
          for (f=0;f<w[e]-'a';++f) {
              i<<=1;
              if (a->info & i)
                 ++j;
          }
          a=a->a[j];
          ++e;
    }

    if (a->info & 1)
       return a->name;
    else
       return -1;
}

#endif

#ifdef MEMORY_INTENSIVE

bool Dictionary::load_dictionary(char *filename) {
     FILE *dict;
     char *s1,*s2,*buffer;
     int e;
     struct AnaObject *an=&anagrams;
     nwords=0;
     buffer=new char [2000000];
     dict=fopen(filename,"r");
     if (dict==NULL)
        return false;
     for (e=0;e<26;++e)
        an->a[e]=NULL;
     an->is_word=false;
     s1=new char [30];
     s2=new char [30];
     while (1) {
        s1=fgets(s1,30,dict);
        if (s1==NULL)
            break;
        s1[strlen(s1)-1]=0;
        s2[0]=0;
        for (e=0;e<strlen(s1);++e)
        {
         s1[e]=tolower(s1[e]);
         if (s1[e]>='a' && s1[e]<='z')
            strncat(s2,s1+e,1);
        }
        if (strlen(s2)>15)               /* 15 letters is limit for Scrabble */
           continue;
        if (strlen(s2)) {
         word[nwords]=buffer;
         strcpy(word[nwords],s2);
         buffer+=strlen(s2)+1;

         for (e=0;e<strlen(s2);++e) {
            if (an->a[s2[e]-'a']==NULL)
                an->a[s2[e]-'a']=create_new_AnaObject();
            an=an->a[s2[e]-'a'];
         }
         an->is_word=true;
         an->name=nwords;
         an=&anagrams;

         ++nwords;
      }
     }
         delete s1;
         delete s2;

     return true;
}

#else

// The memory-conservative version of load_dictionary()

bool Dictionary::load_dictionary(char *filename) {
     FILE *dict;
     char *s1,*s2,*buffer;
     int e,f;
     struct AnaObject *an;
     nwords=0;
     buffer=new char [2000000];
     dict=fopen(filename,"r");
     if (dict==NULL)
        return false;
     anagrams.info=0;
     s1=new char [30];
     s2=new char [30];

     while (1) {
        an=&anagrams;
        s1=fgets(s1,30,dict);
        if (s1==NULL)
            break;
        s1[strlen(s1)-1]=0;
        s2[0]=0;
        for (e=0;e<strlen(s1);++e)
        {
         s1[e]=tolower(s1[e]);
         if (s1[e]>='a' && s1[e]<='z')
            strncat(s2,s1+e,1);
        }
        if (strlen(s2)>15)               /* 15 letters is limit for Scrabble */
           continue;
        if (strlen(s2)) {
         word[nwords]=buffer;
         strcpy(word[nwords],s2);
         buffer+=strlen(s2)+1;

         for (e=0;e<strlen(s2);++e) {
            if (!(an->info & (2 << (s2[e]-'a')))) {
               int i=1,j;
               struct AnaObject **ii;
               j=0;
               for (f=0;f<26;++f) {
                   i<<=1;
                   if (an->info & i)
                      ++j;
               }
               ii=new AnaObject * [j+1];
               for (f=0;f<j;++f)
                   ii[f]=an->a[f];
               ii[f]=create_new_AnaObject();
               if (j>0)
                  delete an->a;
               an->a=ii;
               an->info|=(2 << (s2[e]-'a'));
               an=an->a[f];
            } else {
               int i=1,j=0;
               for (f=0;f<s2[e]-'a';++f) {
                   i<<=1;
                   if (an->info & i)
                      ++j;
               }
               an=an->a[j];
            }
         }
         an->info|=1;
         an->name=nwords;

         ++nwords;
      }
     }

     delete s1;
     delete s2;

     return true;
}

#endif


SBoard::SBoard(int which_board) {
    int e,f;
    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
            contents[e][f]='?';
            is_blank[e][f]=false;
            is_imaginary[e][f]=false;
            shape[e][f]=master_shapes[which_board][e][f];
            bonuses[e][f]=master_bonuses[which_board][e][f];
    }
    is_first_move=true;
    play_defensively=true;
    kill_blank_equity=false;
    moves=new Move [MAX_MOVES];
    their_move=new Move [1000];
    use_feature=true;
    arrowx=-1;
    arrow_across=true;
    two_player_end=false;
    redirect=false;
    return;
}

#ifdef MEMORY_INTENSIVE

void SBoard::valid_anagrams(struct AnaObject *a, bool *oused, Rack *tiles, int call, int needed, int row, int col, int which, int b1, int b2) {

    int e,f,g,n=tiles->n_tiles();
    bool used[7];

    #ifdef USES_TIMERS
     if (redraw_timer)
        update_timer(timer, timer_x, timer_y, false);
    #endif

    if (a==NULL)
        return;
    if ((nmoves==MAX_MOVES && !redirect) || (nmoves==1000 && redirect))
        return;
    if (col>0 && !shape[row][col-1])
        return;

    for (e=0;e<n;++e)
        used[e]=oused[e];
    used[which]=true;

    if (col<15) {

    while (contents[row][col]!='?' && col<=14 && shape[row][col]) {
        a=a->a[contents[row][col]-'a'];
        if (a==NULL)
            return;
        ++col;
    }

    }

    if (a -> is_word && call >= needed)
    {
        if (across) {
            if (!redirect) {
               moves[nmoves].which_word=a->name;
               moves[nmoves].column=col;
               moves[nmoves].row=row;
               moves[nmoves].across=true;
               moves[nmoves].blank[0]=b1;
               moves[nmoves].blank[1]=b2;
            } else {
               their_move[nmoves].which_word=a->name;
               their_move[nmoves].column=col;
               their_move[nmoves].row=row;
               their_move[nmoves].across=true;
               their_move[nmoves].blank[0]=b1;
               their_move[nmoves].blank[1]=b2;
            }
        }
        else {
            if (!redirect) {
               moves[nmoves].which_word=a->name;
               moves[nmoves].column=row;
               moves[nmoves].row=col;
               moves[nmoves].across=false;
               moves[nmoves].blank[0]=b1;
               moves[nmoves].blank[1]=b2;
            } else {
               their_move[nmoves].which_word=a->name;
               their_move[nmoves].column=row;
               their_move[nmoves].row=col;
               their_move[nmoves].across=false;
               their_move[nmoves].blank[0]=b1;
               their_move[nmoves].blank[1]=b2;
            }
        }
        ++nmoves;
    }

    if (col==15)
        return;

    f=0;
    for (e=0;e<n-call;++e) {
        int letter;
        while (f<n && used[f])
            ++f;
        if (f==n)
            return;
        letter=tiles->return_letter(f);
        if (letter==26) {
             for (g=0;g<26;++g) {
                if (legal_letters[row][col][g]) {
                    if (b1!=-1)
                        valid_anagrams(a->a[g],used,tiles,call+1,needed,row,col+1,f,b1,col);
                    else
                        valid_anagrams(a->a[g],used,tiles,call+1,needed,row,col+1,f,col,-1);
                }
             }
        }
        else {
             if (legal_letters[row][col][letter]) {
                bool do_it=true;
                for (g=0;g<f;++g)
                    if (!used[g] && tiles->return_letter(g)==letter) {
                        do_it=false;
                        break;
                    }
                if (do_it)
                    valid_anagrams(a->a[letter],used,tiles,call+1,needed,row,col+1,f,b1,b2);
             }
        }
        ++f;
    }
    return;
}

#else

void SBoard::valid_anagrams(struct AnaObject *a, bool *oused, Rack *tiles, int call, int needed, int row, int col, int which, int b1, int b2) {

    int e,f,g,n=tiles->n_tiles();
    bool used[7];

    #ifdef USES_TIMERS
     if (redraw_timer)
        update_timer(timer, timer_x, timer_y, false);
    #endif

    if (a==NULL)
        return;
    if ((nmoves==MAX_MOVES && !redirect) || (nmoves==1000 && redirect))
        return;
    if (col>0 && !shape[row][col-1])
        return;

    for (e=0;e<n;++e)
        used[e]=oused[e];
    used[which]=true;

    if (col<15) {

    while (contents[row][col]!='?' && col<=14 && shape[row][col]) {
        int i,j;
        if (!(a->info & (2 << (contents[row][col]-'a'))))
           return;
        i=1;
        j=0;
        for (e=0;e<contents[row][col]-'a';++e) {
            i<<=1;
            if (a->info & i)
               ++j;
        }
        a=a->a[j];
        ++col;
    }

    }

    if ((a -> info & 1) && call >= needed)
    {
        if (across) {
            if (!redirect) {
               moves[nmoves].which_word=a->name;
               moves[nmoves].column=col;
               moves[nmoves].row=row;
               moves[nmoves].across=true;
               moves[nmoves].blank[0]=b1;
               moves[nmoves].blank[1]=b2;
            } else {
               their_move[nmoves].which_word=a->name;
               their_move[nmoves].column=col;
               their_move[nmoves].row=row;
               their_move[nmoves].across=true;
               their_move[nmoves].blank[0]=b1;
               their_move[nmoves].blank[1]=b2;
            }
        }
        else {
            if (!redirect) {
               moves[nmoves].which_word=a->name;
               moves[nmoves].column=row;
               moves[nmoves].row=col;
               moves[nmoves].across=false;
               moves[nmoves].blank[0]=b1;
               moves[nmoves].blank[1]=b2;
            } else {
               their_move[nmoves].which_word=a->name;
               their_move[nmoves].column=row;
               their_move[nmoves].row=col;
               their_move[nmoves].across=false;
               their_move[nmoves].blank[0]=b1;
               their_move[nmoves].blank[1]=b2;
            }
        }
        ++nmoves;
    }

    if (col==15)
        return;

    f=0;
    for (e=0;e<n-call;++e) {
        int letter,j,k;
        j=1;
        k=0;
        while (f<n && used[f])
            ++f;
        if (f==n)
            return;
        letter=tiles->return_letter(f);
        if (letter==26) {
             for (g=0;g<26;++g) {
                j<<=1;
                if (a->info & j) {
                   if (legal_letters[row][col][g]) {
                      if (b1!=-1)
                         valid_anagrams(a->a[k],used,tiles,call+1,needed,row,col+1,f,b1,col);
                      else
                          valid_anagrams(a->a[k],used,tiles,call+1,needed,row,col+1,f,col,-1);
                   }
                   ++k;
                }
             }
        }
        else {
             if (legal_letters[row][col][letter]) {
                bool do_it=true;
                for (g=0;g<f;++g)
                    if (!used[g] && tiles->return_letter(g)==letter) {
                        do_it=false;
                        break;
                    }
                if (do_it) {
                   int i=1,j=0;
                   for (g=0;g<letter;++g) {
                       i<<=1;
                       if (a->info & i)
                          ++j;
                   }
                   if ((a->info & (i << 1)))
                      valid_anagrams(a->a[j],used,tiles,call+1,needed,row,col+1,f,b1,b2);
                }
             }
        }
        ++f;
    }
    return;
}

#endif

#ifdef MEMORY_INTENSIVE

void SBoard::find_anagrams(Rack *tiles, int num_tiles, int row, int column) {
    int e,f,n=tiles->n_tiles();
    struct AnaObject *a=&anagrams;
    bool used[7];
    if (!a)
        return;
    while (contents[row][column]!='?' && shape[row][column]) {
        a=a->a[contents[row][column]-'a'];
        if (!a)
            return;
        ++column;
        if (column==15)
           break;
    }
    for (e=0;e<n;++e)
        used[e]=false;
    if (column+num_tiles-1>14)
        return;
    if (column>14)
        return;
    for (e=0;e<n;++e) {
        int letter=tiles->return_letter(e);
        if (letter==26) {
           for (f=0;f<26;++f)
            if (legal_letters[row][column][f])
               valid_anagrams(a->a[f], used, tiles, 1, num_tiles, row, column+1, e, column, -1);
        }
        else {
           if (legal_letters[row][column][letter]) {
            bool do_it=true;
            for (f=0;f<e;++f)
                if (tiles->return_letter(f)==letter) {
                     do_it=false;
                     break;
                }
            if (do_it)
                valid_anagrams(a->a[letter], used, tiles, 1, num_tiles, row, column+1, e, -1, -1);
           }
        }
    }
    return;
}

#else

void SBoard::find_anagrams(Rack *tiles, int num_tiles, int row, int column) {
    int e,f,n=tiles->n_tiles();
    struct AnaObject *a=&anagrams;
    bool used[7];
    if (!a)
        return;
    while (contents[row][column]!='?' && shape[row][column]) {
        if (!(a->info & (2 << (contents[row][column]-'a'))))
            return;
        else {
             int i=1, j=0;
             for (e=0;e<contents[row][column]-'a';++e) {
                 i<<=1;
                 if (a->info & i)
                    ++j;
             }
             a=a->a[j];
        }
        ++column;
        if (column==15)
           break;
    }
    for (e=0;e<n;++e)
        used[e]=false;
    if (column+num_tiles-1>14)
        return;
    if (column>14)
        return;
    for (e=0;e<n;++e) {
        int letter=tiles->return_letter(e);
        if (letter==26) {
           int i=1,j=0;
           for (f=0;f<26;++f) {
            i<<=1;
            if (legal_letters[row][column][f] && (a->info & i))
               valid_anagrams(a->a[j], used, tiles, 1, num_tiles, row, column+1, e, column, -1);
            if (a->info & i)
               ++j;
           }
        }
        else {
           if (legal_letters[row][column][letter]) {
            bool do_it=true;
            for (f=0;f<e;++f)
                if (tiles->return_letter(f)==letter) {
                     do_it=false;
                     break;
                }
            if (do_it && (a->info & (2 << letter))) {
                int i=1, j=0;
                for (f=0;f<letter;++f) {
                    i<<=1;
                    if (a->info & i)
                       ++j;
                }
                valid_anagrams(a->a[j], used, tiles, 1, num_tiles, row, column+1, e, -1, -1);
            }
           }
        }
    }
    return;
}

#endif

int SBoard::with_rack(Rack *tiles, Dictionary *dict, int move) {
        char *w;
        w=dict->report_word(moves[move].which_word);
        if (moves[move].blank[0]!=-1)
                return int(w[moves[move].blank[0]]);
        return 0;
}

bool SBoard::move_uses(int move, int tile, Dictionary *dict) {
        int e,s,a,d,aa,ad;
        char *w;
        bool already;
        already=false;
        if (tile==0)
                return false;
        w=dict->report_word(moves[move].which_word);
        s=strlen(w);
        a=moves[move].column;
        d=moves[move].row;
        aa=moves[move].across ? 1 : 0;
        ad=moves[move].across ? 0 : 1;
        for (e=0;e<s;++e,a+=aa,d+=ad)
                if (contents[d][a]!=w[e]) {
                        if (already)
                                return false;
                        if (int(w[e])!=tile)
                                return false;
                        already=true;
                }
        return true;
}

int SBoard::rid_of(Rack *tiles, Dictionary *dict, int move, struct Fish *fish) {
        int e,s,a,d,aa,ad,c[27],tu=0;
        char *w;
        bool already;
        already=false;
        w=dict->report_word(moves[move].which_word);
        s=strlen(w);
        tiles->copy_rack(c);
        a=moves[move].column;
        d=moves[move].row;
        aa=moves[move].across ? 1 : 0;
        ad=moves[move].across ? 0 : 1;
        for (e=0;e<s;++e,a+=aa,d+=ad)
                if (contents[d][a]!=w[e]) {
                        if (int(contents[d][a])==fish[move].fish_tile && !already)
                                already=true;
                        else
                                --c[w[e]-'a'];
                }
        for (e=0;e<26;++e)
                if (c[e]) {
                        tu=e+'a';
                        break;
                }
        return tu;
}

void SBoard::find_fish_moves(Rack *tiles, Dictionary *dict, Well *well) {

        // Finds moves, and then adjusts their equity according to the
        // probability of drawing a tile which allows a very good move later.

        int c,e;
        struct Fish *fish;

        if (tiles->how_many(26)) {
                // Don't bother if there's a blank on the rack.
                find_moves(tiles, dict, well);
                return;
        }

        tiles->add_letter('?');
        kill_blank_equity=true;
        find_moves(tiles, dict, well);
        tiles->del_letter('?');
        kill_blank_equity=false;

        fish=new struct Fish[nmoves];

        for (e=0;e<nmoves;++e) {
                if (moves[e].row==999 || moves[e].column==999) {
                        fish[e].is_fish=false;
                        continue;
                }
                c=with_rack(tiles,dict,e);
                if (c!=0) {
                        fish[e].is_fish=true;
                        fish[e].fish_tile=c;
                        fish[e].rid=rid_of(tiles,dict,e,fish);
                } else
                        fish[e].is_fish=false;
        }

        for (e=0;e<nmoves;++e) {
                if (!fish[e].is_fish) {
                        bool b[26];
                        int f,g;
                        for (g=0;g<26;++g)
                            b[g]=false;
                        for (f=0;f<500;++f) {
                                if (f==nmoves)
                                        break;
                                if (fish[f].is_fish) {
                                        if (move_uses(e,fish[f].rid,dict) && !b[fish[f].rid-'a']) {
                                                b[fish[f].rid-'a']=true;
                                                moves[e].equity+=int((well->prob(fish[f].fish_tile)*float(moves[f].score + (moves[f].equity/10.)))*10.+.5);
                                        }
                                }
                        }
                }
        }

        c=0;
        for (e=0;e<nmoves;++e)
                if (!fish[e].is_fish)
                        moves[c++]=moves[e];
        nmoves=c;

        delete [] fish;

        qsort((void *)(&moves[0]), nmoves, sizeof(struct Move), compare_scores);

        return;

}

void SBoard::find_moves(Rack *tiles, Dictionary *dict, Well *well) {

    int e,f,g;

    char copy[15][15];
    bool copy_is_blank[15][15],copy_shape[15][15];
    struct Move *mm;

    if (redirect)
       mm=their_move;
    else
        mm=moves;

    across=true;

    nmoves=1;

    // first, let's create the "Pass" and "Exchange" moves

    mm[0].row=999;
    mm[0].column=0;
    mm[0].score=0;
    mm[0].equity=0;                  // "Pass"

    if (well->t_left()>=7) {
       mm[1].row=0;
       mm[1].column=999;
       mm[1].score=0;
       mm[1].equity=150;             // "Exchange"
       ++nmoves;
    }

    if (play_defensively) {

    // Let's compile a list of those tiles still out.

     for (e=0;e<27;++e)
         still_out[e]=letter_dist[e]-tiles->how_many(e);

     for (e=0;e<15;++e)
         for (f=0;f<15;++f) {
             if (contents[e][f]!='?') {
                if (is_blank[e][f])
                    still_out[26]--;
                else
                    still_out[contents[e][f]-'a']--;
             }
         }

     // Initialize open[4].

        init_open();

     }

    // if it's the first move, only find each anagram and place it
    // wherever it may be placed

    if (is_first_move) {
        perform_first_move(tiles, dict, well);
        return;
    }

    while (1) {

    // first, let's determine which are "must include" spaces

    for (e=0;e<15;++e) {
           for (f=0;f<15;++f) {
                    must_include[e][f]=false;
                    if (!shape[e][f])
                        continue;
                    if (contents[e][f]!='?' ||
                           (e>0 && contents[e-1][f]!='?') ||
                           (e<14 && contents[e+1][f]!='?')) {
                       must_include[e][f]=true;
                    }
        }
    }

    // now, let's decide what tiles may be played in what places

    for (e=0;e<15;++e)
        for (f=0;f<15;++f)
                for (g=0;g<26;++g)
                        legal_letters[e][f][g]=true;
    for (e=0;e<15;++e)
         for (f=0;f<15;++f) {
                    if (must_include[e][f] && contents[e][f]=='?') {
                        char word[16], *c;
                        int row=e, column=f, i;

                        word[0]='?';
                        word[1]=0;
                        if (row>0 && contents[--row][column]!='?') {
                        do {
                             for (i=15;i>=1;--i)
                                word[i]=word[i-1];
                             word[0]=contents[row][column];
                             --row;
                        } while (row>=0 && contents[row][column]!='?');
                        }
                        row=e;
                        i=strlen(word);
                        if (row<14 && contents[++row][column]!='?') {
                        do {
                             word[i+1]=0;
                             word[i]=contents[row][column];
                             ++i;
                             ++row;
                        } while (row<=14 && contents[row][column]!='?');
                        }
                        c=strchr(word,'?');
                        for (g=0;g<26;++g) {
                            *c=('a'+g);
                            if (dict->check_word(word)!=-1)
                                legal_letters[e][f][g]=true;
                            else
                                legal_letters[e][f][g]=false;
                        }
                    }
                }
  
    // Now, let's begin.

    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
            int space_ok=8,tiles_used=0;
            if (!shape[e][f])
                continue;
            for (g=f;g<15;++g) {
                if (contents[e][g]=='?')
                        tiles_used++;
                if (must_include[e][g]) {
                        space_ok=tiles_used;
                        break;
                }
                if (tiles_used>7)
                        break;
            }
            if (space_ok>7 || space_ok>tiles->n_tiles())
                continue;

            find_anagrams(tiles,space_ok,e,f);

            if (contents[e][f]!='?') {
                while (f<14 && contents[e][f+1]!='?')
                        ++f;
                ++tiles_used;
                ++f;
            }

        }

    // If we've found the across moves, let's find the down moves.
    // If we've found the down moves, let's break out.

    if (!across)
        break;
    else {
                // Transform the board and repeat

                for (e=0;e<15;++e)
                        for (f=0;f<15;++f) {
                                copy[e][f]=contents[e][f];
                                copy_is_blank[e][f]=is_blank[e][f];
                                copy_shape[e][f]=shape[e][f];
                        }

                for (e=0;e<15;++e)
                        for (f=0;f<15;++f) {
                                contents[e][f]=copy[f][e];
                                is_blank[e][f]=copy_is_blank[f][e];
                                shape[e][f]=copy_shape[f][e];
                        }

                across=false;

    }

    }

    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
                contents[e][f]=copy[e][f];      // Rereflect the board
                is_blank[e][f]=copy_is_blank[e][f];
                shape[e][f]=copy_shape[e][f];
        }

    correct_moves(tiles,dict,well);  // This corrects the position, scores each move, and sorts

    return;
}

void SBoard::move_set(int w, Dictionary *d, Rack *r) {
     int aa=0,ad=0,row=moves[w].row,column=moves[w].column,e;
     char *which=d->report_word(moves[w].which_word);
     if (moves[w].across)
        aa=1;
     else
        ad=1;
     for (e=0;e<strlen(which);++e) {
        if (contents[row][column]=='?') {
            if (moves[w].blank[0]==e || moves[w].blank[1]==e) {
                    is_blank[row][column]=true;
                    r->del_letter('?');
            }
            else
                r->del_letter(which[e]);
            contents[row][column]=which[e];
        }
        else
                contents[row][column]=which[e];
        row+=ad;
        column+=aa;
     }
     is_first_move=false;

     return;
}

void SBoard::move_set_imaginary(int w, Dictionary *d, Rack *r) {
     int aa=0,ad=0,row=moves[w].row,column=moves[w].column,e;
     char *which=d->report_word(moves[w].which_word);
     if (moves[w].across)
        aa=1;
     else
        ad=1;
     for (e=0;e<strlen(which);++e) {
        if (contents[row][column]=='?') {
            if (moves[w].blank[0]==e || moves[w].blank[1]==e) {
                    is_blank[row][column]=true;
                    r->del_letter('?');
            }
            else
                r->del_letter(which[e]);
            is_imaginary[row][column]=true;
            contents[row][column]=which[e];
        }
        else
                contents[row][column]=which[e];
        row+=ad;
        column+=aa;
     }

     return;
}

void SBoard::perform_first_move(Rack *r, Dictionary *d, Well *w) {
    int e,f,g;
    bool copy_shape[15][15];

    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
                copy_shape[e][f]=shape[e][f];
                for (g=0;g<26;++g)
                        legal_letters[e][f][g]=true;
        }

    across=true;
    for (e=1;e<8;++e)
        find_anagrams(r,8-e,7,e);

    across=false;
    for (e=0;e<15;++e)
        for (f=0;f<15;++f)
            shape[e][f]=copy_shape[f][e];
    for (e=1;e<8;++e)
        find_anagrams(r,8-e,7,e);

    for (e=0;e<15;++e)
        for (f=0;f<15;++f)
            shape[e][f]=copy_shape[e][f];

    correct_moves(r,d,w);

    return;

}

void SBoard::correct_moves(Rack *rack, Dictionary *dict, Well *well) {
    int e,l;
    struct Move *mm;

    if (redirect)
       mm=their_move;
    else
        mm=moves;

    for (e=0;e<nmoves;++e) {
        if (mm[e].row==999 || mm[e].column==999)
           continue;           // these denote a "Pass" or "Exchange" move

    #ifdef USES_TIMERS
     if (redraw_timer)
        update_timer(timer, timer_x, timer_y, false);
    #endif

        mm[e].triple=-1;

        if (mm[e].across) {
            l=strlen(dict->report_word(mm[e].which_word));
            mm[e].column-=l;
            if (mm[e].row==0 && mm[e].column==0 && contents[0][0]=='?')
                mm[e].triple=0;
            else if (mm[e].row==0 && mm[e].column<=7 && mm[e].column+l>7 && contents[0][7]=='?')
                mm[e].triple=1;
            else if (mm[e].row==0 && mm[e].column+l==15 && contents[0][14]=='?')
                mm[e].triple=2;
            else if (mm[e].row==7 && mm[e].column==0 && contents[7][0]=='?')
                mm[e].triple=3;
            else if (mm[e].row==7 && mm[e].column+l==15 && contents[7][14]=='?')
                mm[e].triple=4;
            else if (mm[e].row==14 && mm[e].column==0 && contents[14][0]=='?')
                mm[e].triple=5;
            else if (mm[e].row==14 && mm[e].column<=7 && mm[e].column+l>7 && contents[14][7]=='?')
                mm[e].triple=6;
            else if (mm[e].row==14 && mm[e].column+l==15 && contents[14][14]=='?')
                mm[e].triple=7;
        }
        else {
            l=strlen(dict->report_word(mm[e].which_word));
            mm[e].row-=l;
            if (mm[e].column==0 && mm[e].column==0 && contents[0][0]=='?')
                mm[e].triple=0;
            else if (mm[e].column==0 && mm[e].row<=7 && mm[e].row+l>7 && contents[7][0]=='?')
                mm[e].triple=3;
            else if (mm[e].column==0 && mm[e].row+l==15 && contents[14][0]=='?')
                mm[e].triple=5;
            else if (mm[e].column==7 && mm[e].row==0 && contents[0][7]=='?')
                mm[e].triple=1;
            else if (mm[e].column==7 && mm[e].row+l==15 && contents[14][7]=='?')
                mm[e].triple=6;
            else if (mm[e].column==14 && mm[e].row==0 && contents[0][14]=='?')
                mm[e].triple=2;
            else if (mm[e].column==14 && mm[e].row<=7 && mm[e].row+l>7 && contents[7][14]=='?')
                mm[e].triple=4;
            else if (mm[e].column==14 && mm[e].row+l==15 && contents[14][14]=='?')
                mm[e].triple=7;
        }
        mm[e].score=score_move(e,dict);
        if (mm[e].blank[0]!=-1)
                mm[e].blank[0]-= mm[e].across ? mm[e].column : mm[e].row;
        if (mm[e].blank[1]!=-1)
                mm[e].blank[1]-= mm[e].across ? mm[e].column : mm[e].row;
//        if (!well->is_well_empty())
        mm[e].equity=compute_equity(e,rack,dict,well);
    }

    qsort((void *)(&mm[0]), nmoves, sizeof(struct Move), compare_scores);

    if (well->is_well_empty())
        if (two_player_end)
             two_player(rack, dict, orack, well);
        else if (nmoves<50)
             end_game(rack, dict);           // Let's construct an endgame strategy
    else
        if (play_defensively)
           defensive_strategy(rack,dict);

    for (e=nmoves-1;e>=0;--e)
        if (mm[e].score*10+mm[e].equity<0)
                mm[e].equity=-(mm[e].score*10)+1;
        else
                break;

    qsort((void *)(&mm[0]), ((nmoves<100) ? nmoves : 100), sizeof(struct Move), compare_scores);

    if (mm[0].column==999) {

        // Be sure that an exchange really is the best choice here.

        check_exchange(rack,dict);
        qsort((void *)(&mm[0]), nmoves, sizeof(struct Move), compare_scores);

    }

    return;

}

void SBoard::end_game(Rack *rack, Dictionary *d) {
    int e,m=0,mh=0,r[27];
    struct Move *mm;

    if (redirect)
       mm=their_move;
    else
        mm=moves;

    // Simple endgame strategy: once the well is empty, the computer
    // abandons equity calculations and makes a move tree. The tree with
    // the highest total scoring nodes (allowing for unplayed tiles at the
    // end) gets picked.
    //
    // Currently, this is only a one-player endgame strategy

    for (e=0;e<nmoves-1;++e) {
        int points;
        if (mm[e].column==999 || mm[e].row==999)
                continue;
        rack->copy_rack(r);
        remove_letters(r, e, d);
        points=tree_leaf(r, d, e);
        mm[e].equity=points*10;
    }

    return;
}

bool SBoard::remove_letters(int *r, int which, Dictionary *dict) {
    int s,a,d,aa,ad,e;
    char *w;
    struct Move *mm;
    if (redirect)
       mm=their_move;
    else
        mm=moves;
    w=dict->report_word(mm[which].which_word);
    s=strlen(w);
    a=mm[which].column;
    d=mm[which].row;
    aa=mm[which].across ? 1 : 0;
    ad=mm[which].across ? 0 : 1;
    for (e=0;e<s;++e,a+=aa,d+=ad)
            if (contents[d][a]!=w[e]) {
                if (r[w[e]-'a'])
                        --r[w[e]-'a'];
                else {
                        --r[26];
                        if (r[26]==-1)
                                return false;
                }
            }
    return true;
}

int SBoard::tree_leaf(int *r, Dictionary *d, int move) {
        int e,points,nr[27];
        struct Move *mm;

        if (redirect)
           mm=their_move;
        else
            mm=moves;

        points=0;
        for (e=0;e<26;++e)
                points-=letter_val[e]*r[e];

        for (e=move+1;e<nmoves;++e) {
                int f;
                for (f=0;f<27;++f)
                        nr[f]=r[f];
                if (remove_letters(nr,e,d)) {
                        int p=tree_leaf(nr,d,e);
                        if (p>points)
                                points=p;
                }
        }

        return points+mm[move].score;
}

void SBoard::two_player(Rack *mine, Dictionary *dict, Rack *theirs, Well *well) {

        // Two player endgame strategy. Attempts to block mm when the
        // opponent's rack is known.

        int e,f,g,mmoves,board_constraints[15][15][26];

        for (e=0;e<15;++e)
            for (f=0;f<15;++f)
                for (g=0;g<26;++g)
                    board_constraints[e][f][g]=0;

        mmoves=nmoves;
        redirect=true;
        two_player_end=false;
        find_moves(theirs, dict, well);
        redirect=false;
        for (e=0;e<10;++e) {
            if (e==nmoves)
               break;
            if (their_move[e].row==999)
               continue;

            // What letters in what places could block this move?

            int a,d,aa=0,ad=0,l,tiles_on_move;
            char *w;

            a=their_move[e].column;
            d=their_move[e].row;
            if (their_move[e].across)
               aa=1;
            else
                ad=1;
            w=dict->report_word(their_move[e].which_word);
            l=strlen(w);
            tiles_on_move=0;
            for (f=0;f<l;++f,d+=ad,a+=aa) {
                if (contents[d][a]=='?')
                   tiles_on_move++;
                for (g=0;g<26;++g) {
                    if ('a'+g!=w[f]) {
                       board_constraints[d][a][g]+=their_move[e].score;
                       if (their_move[e].score>=40)
                          board_constraints[d][a][f]+=30;
                       if ((w[f]=='q' && orack->how_many('q'-'a')) ||
                           (w[f]=='z' && orack->how_many('z'-'a')))
                          board_constraints[d][a][g]+=50;
                                                        // Oh, no you don't.
                       else if ((w[f]=='j' && orack->how_many('j'-'a')) ||
                           (w[f]=='x' && orack->how_many('x'-'a')));
                          board_constraints[d][a][g]+=40;
                                                        // I don't want you doing that either.
                    }
                }
            }
        }

        for (e=0;e<mmoves;++e) {

            if (moves[e].row==999)
               continue;

            // Find the highest blocking score obtained by placing this
            // word, and adjust the equity accordingly.

            // If this move allows me to go out, and my score is larger
            // than my opponent's, change the equity to a large value -
            // it's a winning move.

            int a,d,aa=0,ad=0,l,tiles_on_move,high_block=0;
            char *w;

            a=moves[e].column;
            d=moves[e].row;
            if (moves[e].across)
               aa=1;
            else
                ad=1;
            w=dict->report_word(moves[e].which_word);
            l=strlen(w);
            tiles_on_move=0;

            for (f=0;f<l;++f,d+=ad,a+=aa) {
                if (contents[d][a]=='?')
                   ++tiles_on_move;
                if (board_constraints[d][a][w[f]-'a']>high_block)
                   high_block=board_constraints[d][a][w[f]-'a'];
            }

            moves[e].equity+=(high_block >> 1); // was high_block

            if (tiles_on_move==mine->n_tiles()) {
               int i=their_score,k=my_score;
               for (f=0;f<orack->n_tiles();++f) {
                   int j=orack->return_letter(f);
                   if (j!=27) {
                      i-=letter_val[j];
                      k+=letter_val[j];
                   }
               }
               if (moves[e].score+k>i)
                  moves[e].equity=9999;         // I win!
               else
                  moves[e].equity=0;            // I don't like it.
            }

        }

        two_player_end=true;
        nmoves=mmoves;

        return;
}

void SBoard::check_exchange(Rack *r, Dictionary *dict) {

        int e,i;
        struct Move *mm;

        if (redirect)
           mm=their_move;
        else
            mm=moves;

        // This function assures that an exchange is only made if it
        // is really the best possible move.
        //
        // Check the rack leave of each scoring move; if one isn't that bad,
        // promote it to best move.

        i=mm[0].equity;

        for (e=1;e<nmoves;++e) {
                int a,d,f,aa,ad,s,score,vowels,consonants,retains,c[27];
                char *w;
                if (mm[e].row==999)
                        continue;
                w=dict->report_word(mm[e].which_word);
                s=strlen(w);
                r->copy_rack(c);
                a=mm[e].column;
                d=mm[e].row;
                aa=mm[e].across ? 1 : 0;
                ad=mm[e].across ? 0 : 1;
                for (f=0;f<s;++f,a+=aa,d+=ad) {
                        if (contents[d][a]!=w[f]) {
                                if (!c[w[f]-'a'])
                                        --c[26];
                                else
                                        --c[w[f]-'a'];
                        }
                }

                // Now let's analyze the rack leave

                score=2;                // if score is positive, play this instead
                vowels=0;
                consonants=0;
                retains=0;

                for (f=0;f<27;++f) {
                        if (c[f]) {
                                if (f==0 || f==4 || f==8 || f==14 || f==20 || f==24)
                                        ++vowels;
                                else
                                        ++consonants;
                                if (f==16)
                                        score-=6;
                                if (f==0 || f==4 || f==8 || f==13 || f==18 || f==19 || f==20)
                                        score+=1;
                                else {
                                        score-=1;
                                        --retains;
                                }
                                if (c[f]>1)
                                        --retains;
                                if (c[f]==2)
                                        score-=1;
                                if (c[f]>=3)
                                        score-=3;
                        }
                }
                if (consonants-vowels<-2 || consonants-vowels>2)
                        score-=3;
                if (consonants==vowels)
                        score+=1;
                if (vowels+consonants<3)
                        score+=3;
                if (vowels+consonants<2)
                        score+=6;
                if (vowels+consonants>4)
                        score-=8;
                if (mm[e].score>=20)
                        score+=3;
                if (mm[e].score<10)
                        score-=10;
                if (score>=0)
                        mm[e].equity+=i+score;
                else if (retains==0)
                        mm[e].equity=999;
                                        // This move results in a tile leave normally obtained by an
                                        // exchange, but scores points doing it.
                else if (retains>-2 && mm[e].score>=30)
                        mm[e].equity=999;
                                        // Close enough.
        }

        return;
}

struct Move *SBoard::report_move(int which) {
       if (redirect)
          return their_move+which;
       else
          return moves+which;
}

int SBoard::score_move(int which, Dictionary *d) {
     int e,i,multiplier=1,l,word=0,otherwords=0,bingo=0;
     char *c;
     struct Move *mm;

     if (redirect)
        mm=their_move;
     else
         mm=moves;

     c=d->report_word(mm[which].which_word);
     l=strlen(c);

     if (mm[which].across) {

        for (i=0,e=mm[which].column;i<l;++i,++e) {
                c=d->report_word(mm[which].which_word);
                if (contents[mm[which].row][e]=='?')
                        ++bingo;
                if (bonuses[mm[which].row][e]>9 && contents[mm[which].row][e]=='?')
                        multiplier*=(bonuses[mm[which].row][e]/10);
                if (contents[mm[which].row][e]=='?' && bonuses[mm[which].row][e] && bonuses[mm[which].row][e]<10 && mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                        word+=letter_val[c[i]-'a']*bonuses[mm[which].row][e];
                else if (mm[which].blank[0]!=e && mm[which].blank[1]!=e && !is_blank[mm[which].row][e])
                        word+=letter_val[c[i]-'a'];
                if (contents[mm[which].row][e]=='?' &&
                    ((mm[which].row>0 && contents[mm[which].row-1][e]!='?') ||
                    (mm[which].row<14 && contents[mm[which].row+1][e]!='?'))) {
                        int mmm=1,r1=mm[which].row-1,r2=mm[which].row+1;
                        if (bonuses[mm[which].row][e]>9)
                                mmm=(bonuses[mm[which].row][e]/10);
                        if (bonuses[mm[which].row][e] && bonuses[mm[which].row][e]<10 && mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                                otherwords+=letter_val[c[i]-'a']*bonuses[mm[which].row][e]*mmm;
                        else if (mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                                otherwords+=letter_val[c[i]-'a']*mmm;
                        while (r1>=0 && contents[r1][e]!='?') {
                                if (!is_blank[r1][e])
                                        otherwords+=letter_val[contents[r1][e]-'a']*mmm;
                                r1--;
                        }
                        while (r2<=14 && contents[r2][e]!='?') {
                                if (!is_blank[r2][e])
                                        otherwords+=letter_val[contents[r2][e]-'a']*mmm;
                                r2++;
                        }
                }
        }

        if (bingo==7) {
                bingo=50;
                mm[which].is_bingo=true;
        }
        else {
                bingo=0;
                mm[which].is_bingo=false;
        }
        word*=multiplier;
        return (word + otherwords + bingo);

     } else {

        for (i=0,e=mm[which].row;i<l;++i,++e) {
                c=d->report_word(mm[which].which_word);
                if (contents[e][mm[which].column]=='?')
                        ++bingo;
                if (bonuses[e][mm[which].column]>9 && contents[e][mm[which].column]=='?')
                        multiplier*=(bonuses[e][mm[which].column]/10);
                if (contents[e][mm[which].column]=='?' && bonuses[e][mm[which].column] && bonuses[e][mm[which].column]<10 && mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                        word+=letter_val[c[i]-'a']*bonuses[e][mm[which].column];
                else if (mm[which].blank[0]!=e && mm[which].blank[1]!=e && !is_blank[e][mm[which].column])
                        word+=letter_val[c[i]-'a'];
                if (contents[e][mm[which].column]=='?' &&
                    ((mm[which].column>0 && contents[e][mm[which].column-1]!='?') ||
                    (mm[which].column<14 && contents[e][mm[which].column+1]!='?'))) {
                        int mmm=1,c1=mm[which].column-1,c2=mm[which].column+1;
                        if (bonuses[e][mm[which].column]>9)
                                mmm=(bonuses[e][mm[which].column]/10);
                        if (bonuses[e][mm[which].column] && bonuses[e][mm[which].column]<10 && mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                                otherwords+=letter_val[c[i]-'a']*bonuses[e][mm[which].column]*mmm;
                        else if (mm[which].blank[0]!=e && mm[which].blank[1]!=e)
                                otherwords+=letter_val[c[i]-'a']*mmm;
                        while (c1>=0 && contents[e][c1]!='?') {
                                if (!is_blank[e][c1])
                                        otherwords+=letter_val[contents[e][c1]-'a']*mmm;
                                --c1;
                        }
                        while (c2<=14 && contents[e][c2]!='?') {
                                if (!is_blank[e][c2])
                                        otherwords+=letter_val[contents[e][c2]-'a']*mmm;
                                c2++;
                        }
                }
        }

        if (bingo==7) {
                bingo=50;
                mm[which].is_bingo=true;
        }
        else {
                bingo=0;
                mm[which].is_bingo=false;
        }
        word*=multiplier;
        return (word + otherwords + bingo);

     }

     return 0;

}

int SBoard::compute_equity(int which, Rack *rack, Dictionary *dict, Well *well) {
    int r_cont[27],e,aa=0,ad=0,a,d,eq,vowels,consonants,tiles_left=0;
    char *w;
    struct Move *mm;

    if (redirect)
       mm=their_move;
    else
        mm=moves;

    eq=0;
    rack->copy_rack(r_cont);
    w=dict->report_word(mm[which].which_word);
    a=mm[which].column;
    d=mm[which].row;
    if (mm[which].across)
        aa=1;
    else
        ad=1;
    for (e=0;e<strlen(w);++e) {
        if (contents[d][a]=='?') {
            if (w[e]=='q' && e!=mm[which].blank[0] && e!=mm[which].blank[1] && rack->how_many('q'-'a'))
               eq+=65+(100-well->t_left());
                                          // equity bonus for using the 'Q'
                                          // (as it becomes more urgent to
                                          // use it, the bonus goes up.)
            if (r_cont[w[e]-'a'])
                r_cont[w[e]-'a']--;
            else
                r_cont[26]--;
        }
        a+=aa;
        d+=ad;
    }
    vowels=0;
    consonants=0;
    for (e=0;e<26;++e) {
        tiles_left+=r_cont[e];
        if (e==0 || e==4 || e==8 || e==14 || e==20)
                vowels+=r_cont[e];
        else
                consonants+=r_cont[e];
        if (r_cont[e]==2 && e!=18) {
                // there's no penalty for doubling 's' or 'e'
                if (e==17 || e==19 || e==8 || e==13)
                        // only a slight penalty for doubling a,i,n,r,t
                        eq-=30;
                else
                        // heavier penalty for doubling something else
                        eq-=40;
        }
        if (r_cont[e]>2) {
                // larger penalities for leaving more than 2 of something
                // on the rack
                if (e==18 || e==4)
                        eq-=40*(r_cont[e]-2);
                else if (e==17 || e==19 || e==8 || e==13)
                        eq-=50*(r_cont[e]-2);
                else
                        eq-=60*(r_cont[e]-2);
        }
        // The only thing which it's O.K. to leave doubles of is the blank :)
        if (r_cont[e]==1) {
                // Let's give equity adjustments for single tiles
                //    (Most of these equity values are based directly on the
                //    number of bingoes possible with that tile.)
                switch (e) {
                        case 0:                 //    'A'
                                eq+=10;
                                break;
                        case 2:
                                eq+=5;          //    'C'
                                break;
                        case 3:
                                eq+=6;          //    'D'
                                break;
                        case 4:                 //    'E'
                                eq+=20;         // was 15
                                break;
                        case 6:
                                eq+=4;          //    'G'
                                break;
                        case 8:                 //    'I'
                                eq+=10;
                                break;
                        case 11:
                                eq+=7;          //    'L'
                                break;
                        case 12:
                                eq+=4;          //    'M'
                                break;
                        case 13:                //    'N'
                                eq+=8;
                                break;
                        case 14:                //    'O'
                                eq+=7;
                                break;
                        case 15:                //    'P'
                                eq+=4;
                                break;
                        case 16:                //    'Q'
                                if ((rack->how_many('u'-'a') || rack->how_many(26)) && !r_cont['u'-'a'] && !r_cont[26])
                                    eq-=125;    // we don't want to leave that Q without a U!
                                break;
                        case 17:                //    'R'
                                eq+=10;
                                break;
                        case 18:                //    'S'
                                eq+=80;         // was 45
                                break;
                        case 19:                //    'T'
                                eq+=8;
                                break;
                        case 20:                //    'U'
                                eq+=6;
                                break;
                        case 21:
                                eq-=1;          //    'V'
                                break;
                        case 23:                //    'X'
                                eq+=80;
                                if (r_cont['a'-'a']+r_cont['i'-'a']+r_cont['c'-'a']+r_cont[26] < 2)
                                     eq-=40;    // 'X' isn't as good without certain tiles
                                break;
                        case 25:                //    'Z'
                                eq+=25;
                                if (!r_cont['i'-'a'] && !r_cont['e'-'a'] && !r_cont[26])
                                   eq-=10;      // the full bonus only applies when we're very likely to score big!
                }
        }
    }

    if (!still_out['u'-'a'] && still_out['q'-'a'] && rack->how_many('u'-'a') &&
       !r_cont['u'-'a'])
       eq-=125;     // Don't let the last u go if there's a q.

    if ( ((vowels-consonants<0) ? consonants-vowels : vowels-consonants) > r_cont[26])
        eq-=75;     // This move leaves an imbalanced rack.


    char *equity_tiles3[]={"ers","eis","aes","aer","eir","est","ert","els",
                       "ein","ens","ars","ael","aet","eos","gin","eor",
                       "eit","ins","eer","eil","elr","der","des","ais",
                       "irs","ees","aen","ain","ist","ern","aer","ast"};
    int equity_by_bingoes3[]={38,31,31,30,28,28,25,24,22,
                           22,22,22,22,22,21,21,21,20,
                           20,19,19,19,19,19,19,19,19,
                           18,18,18,17,17,17,17,17,17};

    for (e=0;e<27;++e)
        if (r_cont[equity_tiles3[e][0]-'a'] &&
            r_cont[equity_tiles3[e][1]-'a'] &&
            r_cont[equity_tiles3[e][2]-'a']) {
                eq+=(equity_by_bingoes3[e]);
                break;
        }

    char *equity_tiles4[]={"aers","eirs","erst","eist","aest","eors","eins",
                       "aert","eers","aels","eirt","eils","aelr","elrs","einr",
                       "aens","aeis","ersu","aeir","agin","ader","eost","enrs",
                       "elst","ders","dire","deis","eprs","ains","eint","erss"};
    int equity_by_bingoes4[]={14,12,12,10,10,10,10,9,9,
                              9,8,8,8,8,8,8,8,8,
                              7,7,7,7,7,7,7,7,7};

    for (e=0;e<27;++e)
        if (r_cont[equity_tiles4[e][0]-'a'] &&
            r_cont[equity_tiles4[e][1]-'a'] &&
            r_cont[equity_tiles4[e][2]-'a'] &&
            r_cont[equity_tiles4[e][3]-'a']) {
                eq+=(equity_by_bingoes4[e]);
                break;
        }

    // Equity bonuses are given out for the following combinations of
    // letters:
    //
    //    "ing"   "re"    "ed"   "ered"   "ers"   "our"   "out"   "ior"
    //    "em"    "man"   "men"  "ish"    "age"   "ion"   "ly"    "mis"
    //    "dis"   "est"   "ier"
    //
    // in the hopes that they might be used in bingoes or super high
    // scoring plays later on!

    int bonus=0;

    if (r_cont['e'-'a']) {
       if (r_cont['s'-'a']) {
          if (r_cont['t'-'a'])
             bonus=50;       // equity bonus for "est" (was 50)
       }
       if (r_cont['r'-'a']) {
          bonus=50;          // equity bonus for "er" or "re" (was 50)
          if (r_cont['s'-'a']) {
             bonus+=25;       // additional equity bonus for "ers" (was 25)
          }
          else if (r_cont['i'-'a']) {
             bonus+=10;       // or for "ier"
          }
       }
       if (r_cont['m'-'a'] && !bonus) {
          bonus=7;           // equity bonus for "em" (was 10)
          if (r_cont['n'-'a']) {
             bonus+=12;       // additional equity bonus for "men" (was 20)
          }
       }
    }

    if (r_cont['i'-'a']) {
       if (r_cont['n'-'a']) {
          if (r_cont['g'-'a'] && bonus<50)
             bonus=50;       // equity bonus for "ing" (was 50)
       }
       if (r_cont['s'-'a'] && bonus<=30) {
          // only one of the following bonuses per rack is allowed
          // ("mise" is good to leave, but not "misdh"!)
          if (r_cont['m'-'a'])
             bonus=20;       // equity bonus for "mis" (was 40)
          else if (r_cont['d'-'a'])
             bonus=20;       // equity bonus for "dis" (was 40)
          else if (r_cont['h'-'a'])
             bonus=30;       // equity bonus for "ish" (was 30)
       }
    }

    if (r_cont['o'-'a']) {
       if (r_cont['i'-'a']) {
          if (r_cont['r'-'a'] && bonus<20)
             bonus=10;       // equity bonus for "ior" (was 20)
          if (r_cont['n'-'a'] && bonus<20)
             bonus=10;       // equity bonus for "ion" (was 20)
       }
       if (r_cont['u'-'a']) {
          if (r_cont['t'-'a'] && bonus<30)
             bonus=10;       // equity bonus for "out" (was 25)
          if (r_cont['r'-'a'] && bonus<20)
             bonus=10;       // equity bonus for "our" (was 20)
       }
    }

    if (r_cont['d'-'a']) {
       if (r_cont['e'-'a'] && bonus<40) {
          bonus=40;          // equity bonus for "ed" or "de" (was 40)
          if (r_cont['e'-'a']>1 && r_cont['r'-'a'])
             bonus+=15;       // additional bonus for "ered"
       }
    }

    if (r_cont['l'-'a'] && bonus<35) {
       if (r_cont['y'-'a'])
          eq+=15;          // equity bonus for "ly"    (was 35)
    }

    eq+=bonus;

    // if three or more nonblank tiles are left, and all or all but one are
    // consonants, take off a few points of equity
    // If there are zero or one consonants in a rack of 4 or more, take off
    // a few points of equity.

    if (tiles_left>=3) {
       int v=r_cont[0]+r_cont[4]+r_cont[8]+r_cont[14]+r_cont[21];
       if (v<=0)
          eq-=(tiles_left-v)*12;
       if (tiles_left>=4 && v>=tiles_left-1)
          eq-=v*12;
    }

    // Turnover: if less than three nonblank tiles remain, a turnover bonus
    // depending on the number of tiles remaining in the well is given

    if (tiles_left<3)
       eq+=(100-well->t_left() >> 2)+20;

    // Blanks shouldn't be used unless using them adds 20 points or more to
    // the score of the play

    if (!kill_blank_equity) {            // used by find_fish_moves

    if (r_cont[26] && tiles_left+well->t_left()>=7) {
        if (r_cont[26]==2)
           eq+=300;
        else
           eq+=200;
    }

    }

    return eq;

}

void SBoard::defensive_strategy(Rack *rack, Dictionary *dict) {
     int a,aa,ad,d,e,f,l,t,which,simple_score,tiles_out=0;
     char *w;
     bool is_left_top,is_either;
     struct Move *mm;

     if (redirect)
        mm=their_move;
     else
         mm=moves;

     // Defensive adjustments are only made for the best 500 plays.

     t=((nmoves<500) ? nmoves : 500);

     for (e=0;e<27;++e)
        tiles_out+=still_out[e];

     for (which=0;which<t;++which) {

     if (mm[which].column==999 || mm[which].row==999)
        continue;

     if (use_feature) {

     danger_spots(which);       // Adjust equity for opening/closing
                                // a Triple Word play.

     }

     aa=0;                      // Initialization
     ad=0;
     simple_score=0;
     is_left_top=false;
     is_either=false;

     w=dict->report_word(mm[which].which_word);
     l=strlen(w);
     if (mm[which].across)
        aa=1;
     else
        ad=1;
     a=mm[which].column;
     d=mm[which].row;

     // First, let's check for Triple Letter or Double Letter spaces
     // opened by this move on both sides. The equity of this move
     // is then adjusted accordingly.

     for (e=0;e<l;++e,a+=aa,d+=ad) {
         int cc1=a,cc2=a,cr1=d,cr2=d;
         char let=w[e];

         simple_score+=letter_val[let-'a'];

         if (let=='k' || let=='j')
            is_left_top=true;
         if (let=='c' || let=='v' || let=='q' || let=='z')
            is_either=true;

         if (contents[d][a]!='?')
            continue;

         if (mm[which].across) {
            --cr1;
            ++cr2;
         } else {
            --cc1;
            ++cc2;
         }
         if (VOWEL(let) && cr1>=0 && cc1>=0 && contents[cr1][cc1]=='?' && bonuses[cr1][cc1]) {
            if (bonuses[cr1][cc1]==3) {
               mm[which].equity-=10;         // Leaving a TL space open
               if (!still_out['x'-'a'])
                  mm[which].equity+=4;       // Oh, well, that isn't so bad.
               if (let=='u' && still_out['q'-'a'])
                  mm[which].equity-=9;       // Now, that's living dangerously!
            }
            if (bonuses[cr1][cc1]==2) {
               mm[which].equity-=5;          // Leaving a DL space open
               if (!still_out['x'-'a'])
                  mm[which].equity+=2;       // As long as the X isn't out...
               if (let=='u' && still_out['q'-'a'])
                  mm[which].equity-=6;
            }
         }
         if (VOWEL(let) && cr2<=14 && cc2<=14 && contents[cr2][cc2]=='?' && bonuses[cr2][cc2]) {
            if (bonuses[cr2][cc2]==3) {
               mm[which].equity-=10;         // Leaving a TL space open
               if (!still_out['x'-'a'])
                  mm[which].equity+=4;       // Oh, well, that isn't so bad.
            }
            if (bonuses[cr2][cc2]==2) {
               mm[which].equity-=5;          // Leaving a DL space open
               if (!still_out['x'-'a'])
                  mm[which].equity+=2;       // As long as the X isn't out...
            }
         }
     }

     // Now let's make sure that the Triple Word lines aren't open.
     // If I'm playing a 'K' or 'J' by the left/top triple line, I'll
     // give myself a pat on the back. If I'm playing a 'C', 'Q' 'V', or
     // 'Z' one away from either triple line, I'll also boost the equity.

     if (mm[which].across) {
        if (mm[which].column==0 &&
           ((mm[which].row>0 && mm[which].row<7 && (contents[0][0]=='?' || contents[7][0]=='?')) ||
            (mm[which].row>7 && mm[which].row<14 && (contents[7][0]=='?' || contents[14][0]=='?')) ))
             // I'm opening a Triple Word Score... naughty boy!
             mm[which].equity-=30;
        if (aa==14 &&
           ((mm[which].row>0 && mm[which].row<7 && (contents[0][14]=='?' || contents[7][14]=='?')) ||
            (mm[which].row>7 && mm[which].row<14 && (contents[7][14]=='?' || contents[14][14]=='?')) ))
             mm[which].equity-=30;
        if (mm[which].row==1 && is_left_top)
             mm[which].equity+=20;     // Let's see you play off of that, meanie!
        else if ((mm[which].row==1 || mm[which].row==14) && is_either)
             mm[which].equity+=20;     // Take that, you crumbum!
     }
     else {
        // The analogous code, playing downwards
        if (mm[which].row==0 &&
           ((mm[which].column>0 && mm[which].column<7 && (contents[0][0]=='?' || contents[0][7]=='?')) ||
            (mm[which].column>7 && mm[which].column<14 && (contents[0][7]=='?' || contents[0][14]=='?')) ))
             mm[which].equity-=30;
        if (ad==14 &&
           ((mm[which].column>0 && mm[which].column<7 && (contents[14][0]=='?' || contents[14][7]=='?')) ||
            (mm[which].column>7 && mm[which].column<14 && (contents[14][7]=='?' || contents[14][14]=='?')) ))
             mm[which].equity-=30;
        if (mm[which].column==1 && is_left_top)
             mm[which].equity+=20;
        else if ((mm[which].column==1 || mm[which].column==14) && is_either)
             mm[which].equity+=20;
     }

     // Front-and-back hooks.
     // We'll determine the chance the opponent has of making a front or
     // back hook to this word, if applicable.

     int nhooks[]={0,0},check[2][2],eq_change;

     if (tiles_out) {

     eq_change=simple_score << 1;

     // Front hooks first.

     if ((mm[which].across && mm[which].column!=0) ||
         (!mm[which].across && mm[which].row!=0)) {
            int cc, cr;
            bool ok=true;
            cc=mm[which].column;
            cr=mm[which].row;
            if (mm[which].across)
               --cc;
            else
               --cr;
            check[0][0]=cr;
            check[0][1]=cc;
            if (mm[which].across) {
               // If the space one to the left and one up/down is occupied,
               // checking front hooks probably isn't necessary.
               if (cr!=0 && contents[cr-1][cc]!='?')
                  ok=false;
               else if (cr!=14 && contents[cr+1][cc]!='?')
                    ok=false;
            }
            else {
               // If the space one up and one to the right/left is occpied,
               // checking front hooks probably isn't necessary.
               if (cc!=0 && contents[cr][cc-1]!='?')
                  ok=false;
               else if (cc!=14 && contents[cr][cc+1]!='?')
                    ok=false;
            }
            if (ok) {
               char new_word[16];
               for (e=1;e<1+l;++e)
                   new_word[e]=w[e-1];
               new_word[e]=0;
               ok=false;
               for (e=0;e<26;++e) {
                   if (still_out[e] || still_out[26]) {
                      new_word[0]='a'+e;
                      if (dict->check_word(new_word)!=-1) {
                         nhooks[0]+=still_out[e];
                         ok=true;
                      }
                   }
               }
               if (ok)
                  nhooks[0]+=still_out[26];
            }
     }

     // Now the back hooks.

     if ((mm[which].across && mm[which].column+l!=15) ||
         (!mm[which].across && mm[which].row+l!=15)) {
            int cc, cr;
            bool ok=true;
            cc=mm[which].column;
            cr=mm[which].row;
            if (mm[which].across)
               cc+=l;
            else
               cr+=l;
            check[1][0]=cr;
            check[1][1]=cc;
            if (mm[which].across) {
               // If the space one to the right and one up/down is occupied,
               // checking front hooks probably isn't necessary.
               if (cr!=0 && contents[cr-1][cc]!='?')
                  ok=false;
               else if (cr!=14 && contents[cr+1][cc]!='?')
                    ok=false;
            }
            else {
               // If the space one up and one to the right/left is occpied,
               // checking front hooks probably isn't necessary.
               if (cc!=0 && contents[cr][cc-1]!='?')
                  ok=false;
               else if (cc!=14 && contents[cr][cc+1]!='?')
                    ok=false;
            }
            if (ok) {
               char new_word[16],c;
               for (e=0;e<l;++e)
                   new_word[e]=w[e];
               new_word[e+1]=0;
               c=e;
               ok=false;
               for (e=0;e<26;++e) {
                   if (still_out[e] || still_out[26]) {
                      new_word[c]='a'+e;
                      if (dict->check_word(new_word)!=-1) {
                         nhooks[1]+=still_out[e];
                         ok=true;
                      }
                   }
               }
               if (ok)
                  nhooks[1]+=still_out[26];
            }
     }

     float prob;

     for (e=0;e<2;++e) {
         prob=float(nhooks[e])*7./float(tiles_out);
         if (prob>.07) {
            int i=bonuses[check[e][0]][check[e][1]];
            mm[which].equity-=eq_change;
            if (i==20)
               mm[which].equity-=eq_change;
            if (i==30)
               mm[which].equity-=(eq_change << 1);
            if (prob>.14) {
               mm[which].equity-=eq_change;
               if (i==20)
                  mm[which].equity-=eq_change;
               if (i==30)
                  mm[which].equity-=(eq_change << 1);
            }
         }
     }

     }

     }

     return;

}

void SBoard::danger_spots(int which) {
        int e,nopen;
        struct Move *mm;

        if (redirect)
           mm=their_move;
        else
            mm=moves;

        // If there are an odd number of corners open, a play is expected
        // to close one of them by using it.
        //
        // If there are no triples open, a play is penalized if it opens one.

        for (e=0,nopen=0;e<3;++e)
                if (open[e])
                        ++nopen;

        if (((nopen >> 1) << 1)!=nopen) {
                switch (mm[which].triple) {
                        case -1:
                                mm[which].equity-=25;
                                ++nopen;
                                break;
                        case 0:
                                if (!open[0]) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 2:
                                if (!open[1]) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 5:
                                if (!open[2]) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 7:
                                if (!open[3]) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 1:
                                if (!(open[0] || open[1]) || mm[which].across==false) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 6:
                                if (!(open[2] || open[3]) || mm[which].across==false) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 3:
                                if (!(open[0] || open[2]) || mm[which].across==true) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                        case 4:
                                if (!(open[1] || open[3]) || mm[which].across==true) {
                                        mm[which].equity-=25;
                                        ++nopen;
                                }
                                break;
                }
        }

        if (!nopen) {

        if ((mm[which].row==0 || mm[which].row==14) && !mm[which].across) {
                if (mm[which].column<7) {
                        bool ok=false;
                        for (e=mm[which].column;e>=0;e--)
                                if (contents[mm[which].row][e]!='?') {
                                        ok=true;
                                        break;
                                }
                        for (e=mm[which].column;e<=8;e++)
                                if (contents[mm[which].row][e]!='?') {
                                        ok=true;
                                        break;
                                }
                        if (!ok)
                                mm[which].equity-=100;
                }
                else {
                        bool ok=false;
                        for (e=mm[which].column;e<15;e++)
                                if (contents[mm[which].row][e]!='?') {
                                        ok=true;
                                        break;
                                }
                        for (e=mm[which].column;e>=6;e--)
                                if (contents[mm[which].row][e]!='?') {
                                        ok=true;
                                        break;
                                }
                        if (!ok)
                                mm[which].equity-=100;
                }
        }
        if ((mm[which].column==0 || mm[which].column==14) && mm[which].across) {
                if (mm[which].row<7) {
                        bool ok=false;
                        for (e=mm[which].row;e>=0;e--)
                                if (contents[e][mm[which].column]!='?') {
                                        ok=true;
                                        break;
                                }
                        for (e=mm[which].row;e<=8;e++)
                                if (contents[e][mm[which].column]!='?') {
                                        ok=true;
                                        break;
                                }
                        if (!ok)
                                mm[which].equity-=100;
                }
                else {
                        bool ok=false;
                        for (e=mm[which].row;e<15;e++)
                                if (contents[e][mm[which].column]!='?') {
                                        ok=true;
                                        break;
                                }
                        for (e=mm[which].row;e>=6;e--)
                                if (contents[e][mm[which].column]!='?') {
                                        ok=true;
                                        break;
                                }
                        if (!ok)
                                mm[which].equity-=100;
                }
        }

        }

        return;
}

void SBoard::init_open() {
        int e;

        // Are the corner triple word squares open?

        for (e=0;e<4;++e)
                open[e]=false;

        if (contents[0][0]=='?')
                for (e=1;e<6;++e) {
                        if (contents[0][e]!='?' && contents[0][e+1]=='?') {
                                open[0]=true;
                                break;
                        }
                        if (contents[e][0]!='?' && contents[e+1][0]=='?') {
                                open[0]=true;
                                break;
                        }
                }
        if (contents[0][14]=='?')
                for (e=13;e>8;--e) {
                        if (contents[0][e]!='?' && contents[0][e-1]=='?') {
                                open[1]=true;
                                break;
                        }
                        if (contents[14-e][0]!='?' && contents[15-e][0]=='?') {
                                open[1]=true;
                                break;
                        }
                }
        if (contents[14][0]=='?')
                for (e=1;e<6;++e) {
                        if (contents[14][e]!='?' && contents[14][e+1]=='?') {
                                open[2]=true;
                                break;
                        }
                        if (contents[14-e][0]!='?' && contents[13-e][0]=='?') {
                                open[2]=true;
                                break;
                        }
                }
        if (contents[14][14]=='?')
                for (e=13;e<8;--e) {
                        if (contents[14][e]!='?' && contents[14][e-1]=='?') {
                                open[3]=true;
                                break;
                        }
                        if (contents[e][0]!='?' && contents[e-1][0]=='?') {
                                open[3]=true;
                                break;
                        }
                }
        return;
}

void SBoard::set_on_board(int row, int col, char what, int blank) {

    // Used only by the "Scrabble Companion" version driven by this engine

    contents[row][col]=what;
    if (blank)
        is_blank[row][col]=true;
    else
        is_blank[row][col]=false;
    is_first_move=false;
}

void SBoard::dump_board() {

    int e,f;

    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
            contents[e][f]='?';
            is_blank[e][f]=false;
        }
    is_first_move=true;

    return;

}

void SBoard::clear_board(Well *where_to) {
    int e,f;
    for (e=0;e<15;++e)
        for (f=0;f<15;++f) {
            if (contents[e][f]!='?') {
                if (is_blank[e][f])
                        where_to->return_to_well('?');
                else
                        where_to->return_to_well(contents[e][f]);
                is_blank[e][f]=false;
                contents[e][f]='?';
            }
        }
    is_first_move=true;
    return;
}

Rack::Rack() {
    int e;
    ntiles=0;
    for (e=0;e<27;++e)
        c[e]=0;
    for (e=0;e<7;++e)
        display[e]=' ';
    return;
}

int Rack::return_letter(int which_one) {
    int e;
    for (e=0;e<27;++e) {
        which_one-=c[e];
        if (which_one<0)
            return e;
    }
    return 0;
}

void Rack::fill_from_well(Well *w) {
    while (ntiles<7 && !w->is_well_empty()) {
        char cd=w->draw_from_well();
        add_letter(cd);
    }
    return;
}

void Rack::dump_rack() {

    int e;

    for (e=0;e<27;++e)
        c[e]=0;
    ntiles=0;

    return;

}

void Rack::add_letter(char cc) {

    int e;

    cc=tolower(cc);

    if (cc=='?')
        ++c[26];
    else
        ++c[cc-'a'];
    ++ntiles;
    for (e=0;e<7;++e)
        if (display[e]==' ') {
           display[e]=cc;
           break;
        }

    return;
}

bool Rack::del_letter(char cc) {
    int e,f,i;
    cc=tolower(cc);
    i=cc-'a';
    if (cc=='?')
        i=26;
    if (!c[i]) {
        if (c[26])
            --c[26];
        for (e=0;e<7;++e)
            if (display[e]=='?') {
               display[e]=' ';
               break;
            }
        --ntiles;
        return false;
    }
    --c[i];
    for (e=0;e<7;++e)
        if (display[e]==cc) {
           display[e]=' ';
           break;
        }
    --ntiles;
    return true;
}

void Rack::remove_letter(int which) {
     if (display[which]==' ')
        return;
     if (display[which]=='?')
         --c[26];
     else
         --c[tolower(display[which])-'a'];
     display[which]=' ';
     --ntiles;
     return;
}

void Rack::inc(char ch) {
     ch=tolower(ch);
     if (ch=='?')
        ++c[26];
     else
         ++c[ch-'a'];
     ++ntiles;
     return;
}

void Rack::copy_rack(int *i) {
    int e;
    for (e=0;e<27;++e)
        i[e]=c[e];
    return;
}

void Rack::empty_rack(Well *well) {
    int e,f;
    for (e=0;e<26;++e) {
        for (f=0;f<c[e];++f)
                well->return_to_well('a'+e);
        c[e]=0;
    }
    for (e=0;e<c[26];++e)
         well->return_to_well('?');
    c[26]=0;
    ntiles=0;
    for (e=0;e<7;++e)
        display[e]=' ';
    return;

}

void Rack::print_rack(ostream& o) {
    for (int e=0;e<ntiles;++e)
        o << display[e] << " ";
    o << endl;
    return;
}

int Rack::exchange_tiles(Well *w) {

    // Exchanges tiles for the computer player.
    Rack nr;
    int e,i;
    for (e=0;e<26;++e) {
        if (c[e] && ((e!=0 && e!=4 && e!=8 && e!=13 && e!=17 && e!=18 && e!=19 && e!=20) || (e==20 && (w->t_left()>20 || w->prob('q')==0.)))) {
                int f;
                for (f=0;f<c[e];++f) {
                        nr.add_letter('a'+e);
                        del_letter('a'+e);
                }
        }
        else {
                int f;
                for (f=1;f<c[e];++f) {
                        nr.add_letter('a'+e);
                        del_letter('a'+e);
                }
        }
    }
    if (!nr.n_tiles()) {
        // In this freakish emergence, exchange everything but the blank(s).
        int e;
        for (e=0;e<26;++e) {
                int f;
                for (f=0;f<c[e];++f) {
                        nr.add_letter('a'+e);
                        del_letter('a'+e);
                }
        }
    }
    fill_from_well(w);
    i=nr.n_tiles();
    nr.empty_rack(w);

    return i;

}

char *Rack::exchange_string(char *buffer, Well *well) {
    Rack nr;
    int e,i;
    for (e=0;e<8;++e)
        buffer[e]=0;
    for (e=0;e<26;++e) {
        if (c[e] && ((e!=0 && e!=4 && e!=8 && e!=13 && e!=17 && e!=18 && e!=19 && e!=20) || (e==20 && (well->t_left()>20 || well->prob('q')==0.)))) {
                int f;
                for (f=0;f<c[e];++f)
                        nr.add_letter('a'+e);
        }
        else {
                int f;
                for (f=1;f<c[e];++f)
                        nr.add_letter('a'+e);
        }
    }
    if (!nr.n_tiles()) {
        int e;
        for (e=0;e<26;++e) {
                int f;
                for (f=0;f<c[e];++f)
                        nr.add_letter('a'+e);
        }
    }
    for (e=0;e<nr.n_tiles();++e) {
        i=nr.return_letter(e);
        buffer[e]='a'+i;
    }
    return buffer;
}

/* well, well! */

Well::Well() {
    int c=0,e,f;
    srand(time(NULL));
    tiles_left=100;
    for (e=0;e<26;++e)
        for (f=0;f<letter_dist[e];++f)
             contents[c++]='a'+e;
    for (f=0;f<letter_dist[26];++f)
        contents[c++]='?';
    return;
}

void Well::refill_well() {
    int c=0,e,f;
    tiles_left=100;
    for (e=0;e<26;++e)
        for (f=0;f<letter_dist[e];++f)
             contents[c++]='a'+e;
    for (f=0;f<letter_dist[26];++f)
        contents[c++]='?';
    return;
}

bool Well::is_well_empty() {
    return (!tiles_left);
}

char Well::draw_from_well() {
    int i;
    char c;
    if (is_well_empty())
        return '\000';
    do {
        i=rand() % 100;
    } while (contents[i]==' ');
    c=contents[i];
    contents[i]=' ';
    --tiles_left;
    return c;
}

void Well::return_to_well(char cc) {
    int e;
    if (!isalpha(cc) && cc!='?')
        return;
    for (e=0;e<100;++e)
        if (contents[e]==' ') {
               contents[e]=cc;
               tiles_left++;
               return;
        }
}

bool Well::remove_from_well(int letter) {
        bool b=false;
        int e;
        letter=tolower(letter);
        for (e=0;e<100;++e) {
            if (contents[e]==letter) {
               contents[e]=' ';
               b=true;
               --tiles_left;
               break;
            }
        }
        return b;
}

float Well::prob(int c) {
    int e,t;
    t=0;
    if (tiles_left==0)
       return 0.0;
    for (e=0;e<100;++e)
        if (contents[e]==char(c) || contents[e]=='?')
                ++t;
    return float(t)/float(tiles_left);
}

